import pydbr

def test_connect_returns_connection(pydbr_fix):
    dbc = pydbr.connect(
        bearer_token=pydbr_fix.dastabricks_token,
        url=pydbr_fix.databricks_url,
        cluster_id=pydbr_fix.databricks_cluster_id)

    assert isinstance(dbc, pydbr.connection.DatabricksConnection)

def test_connect_with_trailing_slash_for_url_removes_slash(pydbr_fix):
    dbc = pydbr.connect(
        bearer_token=pydbr_fix.dastabricks_token,
        url=pydbr_fix.databricks_url + '/',
        cluster_id=pydbr_fix.databricks_cluster_id)

    assert isinstance(dbc, pydbr.connection.DatabricksConnection)
    assert dbc.link.url == pydbr_fix.databricks_url
 

def test_connection_property_link_is_a_Link(dbc):
    assert hasattr(dbc, 'link')
    assert isinstance(dbc.link, pydbr.common.Link)

def test_connection_property_worksapce_is_a_Workspace(dbc):
    assert hasattr(dbc, 'workspace')
    assert isinstance(dbc.workspace, pydbr.workspace.Workspace)

def test_connection_property_dbfs_is_a_DBFS(dbc):
    assert hasattr(dbc, 'dbfs')
    assert isinstance(dbc.dbfs, pydbr.dbfs.DBFS)

def test_connection_property_jobs_is_a_Jobs(dbc):
    assert hasattr(dbc, 'jobs')
    assert isinstance(dbc.jobs, pydbr.jobs.Jobs)




import requests
from .common import Link
from .workspace import Workspace
from .jobs import Jobs
from .dbfs import DBFS


class DatabricksConnection:
    def __init__(self, bearer_token, url=None, cluster_id=None):
        self._link = Link(bearer_token, url, cluster_id)

    @property
    def link(self) -> Link:
        return self._link

    @property
    def workspace(self) -> Workspace:
        '''Databricks Workspace REST API.'''
        if not getattr(self, "_workspace", None):
            self._workspace = Workspace(self.link)
        return self._workspace

    @property
    def jobs(self) -> Jobs:
        '''Databricks Jobs REST API.'''
        if not getattr(self, "_jobs", None):
            self._jobs = Jobs(self.link)
        return self._jobs

    @property
    def dbfs(self) -> DBFS:
        '''Databricks DBFS REST API.'''
        if not getattr(self, "_dbfs", None):
            self._dbfs = DBFS(self.link)
        return self._dbfs

def connect(bearer_token, url=None, cluster_id=None) -> DatabricksConnection:
    return DatabricksConnection(bearer_token, url, cluster_id)




import os
from dataclasses import dataclass
import pytest
from pyspark.sql import SparkSession
import pydbr
from pydbr.connection import DatabricksConnection

@dataclass
class PydbrTestFixture:
    ENV_VAR_DATABRICKS_TOKEN: str = 'DATABRICKS_BEARER_TOKEN'
    ENV_VAR_DATABRICKS_URL: str = 'DATABRICKS_URL'
    ENV_VAR_DATABRICKS_CLUSTER_ID: str = 'DATABRICKS_CLUSTER_ID'

    dastabricks_token: str = None
    databricks_url: str = None
    databricks_cluster_id: str = None

    dbfs_tmp = '/pydbr-test/tmp'
    dbfs_fix = '/pydbr-test/fixture'
    dbfs_fix_files = '/pydbr-test/fixture/files'

    def __post_init__(self):
        self.databricks_url = os.environ[self.ENV_VAR_DATABRICKS_URL]
        self.dastabricks_token = os.environ[self.ENV_VAR_DATABRICKS_TOKEN]
        self.databricks_cluster_id = os.environ[self.ENV_VAR_DATABRICKS_CLUSTER_ID]


@pytest.fixture(scope="session")
def pydbr_fix():
    """Returns Pydbr test fixture."""
    fixture = PydbrTestFixture()
    yield fixture

@pytest.fixture
def dbc(pydbr_fix) -> DatabricksConnection:
    """Databricks connection fixture."""
    connection = pydbr.connect(
        bearer_token=pydbr_fix.dastabricks_token,
        url=pydbr_fix.databricks_url,
        cluster_id=pydbr_fix.databricks_cluster_id)
    yield connection

@pytest.fixture(scope="session")
def spark():
    import os
    os.environ['SPARK_LOCAL_IP'] = '127.0.0.1'
    spark_session = (SparkSession
        .builder
        .appName('TestingPysparkMe')
        .master('local[*]')
        .enableHiveSupport()
        .getOrCreate())
    spark_session.sparkContext.setLogLevel('ERROR')
    yield spark_session
    try:
        spark_session.stop()
    except:
        pass




import time
import pytest
import pydbr

def test_dbfs_ls(dbc, pydbr_fix):
    actual = dbc.dbfs.ls(pydbr_fix.dbfs_fix_files)
    assert isinstance(actual, list)
    assert 1, len(actual)

    hello_info = actual[0]
    assert isinstance(hello_info, pydbr.databricks_data_classes.DatabricksFileInfo)
    assert f'{pydbr_fix.dbfs_fix_files}/hello.txt' == hello_info.path


def test_dbfs_info(dbc, pydbr_fix):
    actual = dbc.dbfs.info(pydbr_fix.dbfs_fix_files)
    assert isinstance(actual, pydbr.databricks_data_classes.DatabricksFileInfo)
    assert pydbr_fix.dbfs_fix_files == actual.path
    assert True == actual.is_dir
 
def test_dbfs_exists_for_existing_directory_returns_true(dbc, pydbr_fix):
    actual = dbc.dbfs.exists(f'{pydbr_fix.dbfs_fix_files}')
    assert True == actual

def test_dbfs_exists_for_existing_file_returns_true(dbc, pydbr_fix):
    actual = dbc.dbfs.exists(f'{pydbr_fix.dbfs_fix_files}/hello.txt')
    assert True == actual

def test_dbfs_exists_for_missing_file_returns_false(dbc, pydbr_fix):
    actual = dbc.dbfs.exists(f'{pydbr_fix.dbfs_fix_files}/who-am-i.txt')
    assert False == actual


def test_dbfs_read_small_file_returns_file_bytes(dbc, pydbr_fix):
    actual = dbc.dbfs.read(f'{pydbr_fix.dbfs_fix_files}/hello.txt')
    assert actual == b'Hello pydbr!'

def test_dbfs_read_missing_file_raises_exception(dbc, pydbr_fix):
    with pytest.raises(pydbr.common.DatabricksLinkException) as excinfo:
        actual = dbc.dbfs.read(f'{pydbr_fix.dbfs_fix_files}/who-am-i.txt')
    assert excinfo.value.error_code == 'RESOURCE_DOES_NOT_EXIST'


def test_dbfs_read_with_offset_and_lengt_returns_chunk(dbc, pydbr_fix):
    actual = dbc.dbfs.read(f'{pydbr_fix.dbfs_fix_files}/hello.txt',
            offset=3, length=2)
    assert actual == b'lo'

def test_dbfs_read_all_returns_file_bytes(dbc, pydbr_fix):
    actual = dbc.dbfs.read_all(f'{pydbr_fix.dbfs_fix_files}/hello.txt')
    assert actual == b'Hello pydbr!'


def test_dbfs_mkdirs_creates_directories(dbc, pydbr_fix):
    path = f'{pydbr_fix.dbfs_tmp}/some/dir/{pydbr.common.random_id(5, "pydbr-"+str(time.time()))}'    
    actual = dbc.dbfs.mkdirs(path)
    
    assert actual == {}
    assert dbc.dbfs.exists(path) == True

def test_rm_existing_file_path_deletes_it(dbc, pydbr_fix):
    path = f'{pydbr_fix.dbfs_tmp}/some/dir/{pydbr.common.random_id(5, "pydbr-"+str(time.time()))}'    
    dbc.dbfs.mkdirs(path)
    assert dbc.dbfs.exists(path) == True
    
    actual = dbc.dbfs.rm(f'{pydbr_fix.dbfs_tmp}', recursive=True)
    assert dbc.dbfs.exists(pydbr_fix.dbfs_tmp) == False



import enum
import base64
from typing import List
from .common import Api, Link, DatabricksLinkException, ERR_RESOURCE_DOES_NOT_EXIST
from .databricks_data_classes import *

class ExportFormat(enum.Enum):
    SOURCE = 'SOURCE',
    HTML = 'HTML',
    JUPYTER = 'JUPYTER',
    DBC = 'DBC'

class Workspace(Api):
    def __init__(self, link):
        super().__init__(link, path='workspace')

    def ls(self, path=None) -> List[DatabricksObjectInfo]:
        response = self.link.get(
            self.path('list'),
            params=dict(path=(path or '/')))
        objects = [DatabricksObjectInfo(**obj) for obj in response.get('objects', [])]
        return objects

    def exists(self, path):
        try:
            self.ls(path)
            result = True
        except DatabricksLinkException as exc:
            if exc.error_code == ERR_RESOURCE_DOES_NOT_EXIST:
                result = False
        return result        

    def is_directory(self, path):
        if path == '/':
            return True
        item = self.ls(path)[0]
        return item.path != path

    def export(self, path:str, format:ExportFormat=ExportFormat.SOURCE) -> bytes:
        format = ExportFormat[format] if isinstance(format, str) else format
        assert isinstance(format, ExportFormat), 'format must be ExportFormat or str instance'
        link_response = self.link.get(
            self.path('export'),
            params=dict(path=path,
                        format=format.name))
        # return link_response
        content = base64.b64decode(link_response['content'])
        return content




from .common import Api
from .runs import Runs

class Jobs(Api):
    def __init__(self, link):
        super().__init__(link, path='jobs')

    def list(self, path=None):
        list_result = self.link.get(
            self.path('list'),
            params=dict(path=(path or '/')))
        return list_result.get('jobs', [])

    @property
    def runs(self) -> Runs:
        if not getattr(self, "_runs", None):
            self._runs = Runs(self.link)
        return self._runs



import base64
import pytest
import pydbr

def test_link_property_bearer_token_returns_correct_value(dbc, pydbr_fix):
    assert dbc.link.bearer_token == pydbr_fix.dastabricks_token

def test_link_property_url_returns_correct_value(dbc, pydbr_fix):
    assert dbc.link.url == pydbr_fix.databricks_url

def test_link_property_cluster_id_correct_value(dbc, pydbr_fix):
    assert dbc.link.cluster_id == pydbr_fix.databricks_cluster_id


def test_link_get_returns_parsed_response(dbc):
    response = dbc.link.get(verb='dbfs/list', params={'path':'/'})
    assert isinstance(response, dict)
    assert 'files' in response

def test_link_get_with_bad_params_raises_exception(dbc):
    with pytest.raises(pydbr.common.DatabricksLinkException) as excinfo:
        response = dbc.link.get(verb='dbfs/list')
    assert excinfo.value.error_code == 'INVALID_PARAMETER_VALUE'


def test_link_post_with_json_returns_parsed_response(dbc, pydbr_fix):
    response = dbc.link.post(verb='dbfs/put', json={'path':f'{pydbr_fix.dbfs_tmp}/test_link_put.txt',
                                                     'contents': base64.b64encode(b'Hello, pydbr!').decode(), 
                                                     'overwrite': True,
                                                    },
                             params={})
    assert isinstance(response, dict)
    assert {} == response


def test_link_put_with_basd_params_raises_exception(dbc, pydbr_fix):
    
    # Missing contents parameter
    with pytest.raises(pydbr.common.DatabricksLinkException) as excinfo:
        response = dbc.link.post(verb='dbfs/put', json={'path':f'{pydbr_fix.dbfs_tmp}/test_link_put.txt',
                                                        #  'contents': base64.b64encode(b'Hello, pydbr!').decode(), 
                                                        'overwrite': True})
    
    assert excinfo.value.error_code == 'INVALID_PARAMETER_VALUE'


def test_api_has_property_link(dbc):
    api = pydbr.common.Api(dbc.link, 'dbfs')
    assert hasattr(api, 'link')
    assert isinstance(api.link, pydbr.common.Link)

def test_api_path_returns_path_to_action(dbc):
    api = pydbr.common.Api(dbc.link, 'dbfs')
    actual = api.path('list')
    assert 'dbfs/list' == actual


def test_bite_size_str_with_default_format():
    actual = pydbr.common.bite_size_str(100)
    assert '100 B' == actual
    actual = pydbr.common.bite_size_str(2500)
    assert '2.5 KB' == actual

def test_bite_size_str_with_custom_format():
    fmt = '{0:.3n} {1}pix'
    actual = pydbr.common.bite_size_str(200, fmt)
    assert '200 pix' == actual
    actual = pydbr.common.bite_size_str(3500, fmt)
    assert '3.5 Kpix' == actual



import bisect
import requests
import dataclasses
import uuid

DEFAULT_URL = 'https://westeurope.azuredatabricks.net'

ERR_RESOURCE_DOES_NOT_EXIST = 'RESOURCE_DOES_NOT_EXIST'

class Link:
    def __init__(self, bearer_token, url:str=None, cluster_id=None):
        url or DEFAULT_URL
        if url.endswith('/'):
            url = url[:-1]
        self._bearer_token = bearer_token
        self._url = url or DEFAULT_URL
        self._cluster_id = cluster_id

    @property
    def bearer_token(self) -> str:
        '''Databricks bearer access token.'''
        return self._bearer_token

    @property
    def url(self) -> str:
        '''Databricks URL.'''
        return self._url

    @property
    def cluster_id(self) -> str:
        '''Databricks cluster ID.'''
        return self._cluster_id

    def _get_url(self, verb):
        '''Get endpoint URL for a verb.'''
        url = f'{self.url}/api/2.0/{verb}'
        return url

    def _get_headers(self):
        '''Get common Databricks REST API headers.'''
        return dict(
            Authorization = f'Bearer {self.bearer_token}'
        )


    def get(self, verb, params=None):
        '''Execute GET request and return JSON parsed response.'''
        response_obj = requests.get(
                url=self._get_url(verb),
                params=(params or {}),
                headers=self._get_headers()
            )
        response = response_obj.json()
        if response.get('error_code', None):
            raise DatabricksLinkException(response['error_code'], response['message'])
        return response

    def post(self, verb, params=None, json=None):
        '''Execute POST request and return JSON parsed response.'''
        args = dict(
            url=self._get_url(verb),
            headers=self._get_headers()
        )
        if params:
            args['params'] = params
        if json:
            args['headers']['Content-Type'] = 'application/json'
            args['json'] = json
        response_obj = requests.post(**args)
        response = response_obj.json()
        if response.get('error_code', None):
            raise DatabricksLinkException(response['error_code'], response['message'])
        return response



class Api:
    '''Generic API class'''
    def __init__(self, link: Link, path=None):
        self._link = link
        self._path = path or ''
   
    @property
    def link(self) -> Link:
        return self._link

    def path(self, action):
        "Get path to a resource, relative to the root_path"
        action_path = "{}/{}".format(self._path, action)
        return action_path

class DatabricksLinkException(Exception):
    def __init__(self, error_code, message):
        self.error_code = error_code
        self.message = message
        super().__init__('{}: {}'.format(error_code, message))


def bite_size_str(size, format_string=None, base=1000):
    sizes = [base ** 0, base**1, base**2, base**3, base**4,
             base**5, base**6, base**7, base**8]
    prefixes = ['', 'K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y']
    i = bisect.bisect(sizes, size)
    i = 0 if (i<1) else i-1
    number_part = size/sizes[i]
    format_string = format_string or "{0:.3n} {1}B" 
    return format_string.format(size/sizes[i], prefixes[i])


def random_id(length, namespace=''): # pragma: no cover
    namespace_url = 'https://pypi.org/project/pydbr/#' + namespace
    return str(uuid.uuid5(uuid.NAMESPACE_URL, namespace_url)).replace('-','')[:length]




import datetime
from typing import List
from .common import Api
from .databricks_data_classes import *


class Runs(Api):
    def __init__(self, link):
        super().__init__(link, path='jobs/runs')

    def get(self, run_id) -> DatabricksRun:
        response = self.link.get(
            self.path('get'),
            params=dict(run_id=run_id),)
        return DatabricksRun(**response)

    def get_output(self, run_id) -> DatabricksRun:
        response = self.link.get(
            self.path('get-output'),
            params=dict(run_id=run_id),)
        return DatabricksRunOutput.from_dict(response)

    def export(self, run_id) -> DatabricksRun:
        response = self.link.get(
            self.path('export'),
            params=dict(run_id=run_id, views_to_export='ALL',),
            )
        return DatabricksViewList.from_dict(response)

    def submit_notebook(self, path, params=None, run_name=None, cluster_id=None):
        cluster_id = cluster_id or self.link.cluster_id
        assert cluster_id, f"cluster_id not specified. Set cluster_id with connect or pass as parameter"

        params = params or {}
        run_name = run_name or "pyspark-me-" + str(int(datetime.datetime.now().timestamp()))

        r = dict(
            run_name=run_name,
            existing_cluster_id=cluster_id,
            libraries=[],
            notebook_task=dict(
                notebook_path=path,
                base_parameters=params
            ),
        )

        response = self.link.post(
            self.path('submit'),
            json=r
        )
        return response['run_id']

    def ls(self, job_id=None, offset=None, limit=None,
            completed_only=False, active_only=False) -> DatabricksRunList:
        assert not (completed_only and active_only), "Only one of completed_only or active_only could be True"
        params = dict()
        if job_id:
            params['job_id'] = job_id
        if offset:
            params['offset'] = offset
        if limit:
            params['limit'] = limit
        if completed_only:
            params['completed_only'] = 'true'
        if active_only:
            params['active_only'] = 'true'

        response = self.link.get(
            self.path('list'),
            params=params
        )
        return DatabricksRunList(
                runs=[DatabricksRun(**run) for run in response.get('runs',[])],
                has_more=response.get('has_more', False),)



from dataclasses import dataclass, field, asdict
from typing import List
from .common import bite_size_str

class DataClass:
    def asdict(self) -> dict:
        return asdict(self)

    def __iter__(self):
        return iter(self.asdict().items())

    @classmethod
    def from_dict(cls, d:dict):
        return cls(**d)

@dataclass
class DatabricksObjectInfo(DataClass):
    object_type: str
    object_id: str
    path: str
    language: str = None
    is_directory: bool = None
    is_notebook: bool = None
    is_library: bool = None

    def __post_init__(self):
        self.is_notebook = self.object_type == 'NOTEBOOK'
        self.is_directory = self.object_type == 'DIRECTORY'
        self.is_library = self.object_type == 'LIBRARY'


@dataclass
class DatabricksFileInfo(DataClass):
    path: str
    is_dir: bool
    file_size: int
    is_file: bool = None
    human_size: str = field(init=False, compare=False)

    def __post_init__(self):
        self.human_size = bite_size_str(self.file_size)
        self.is_file = not self.is_dir



@dataclass
class DatabricksRun(DataClass):
    job_id: int
    run_id: int
    creator_user_name: str = None
    number_in_job: int = None
    original_attempt_run_id: int = None
    state: dict = None
    schedule: dict = None
    task: dict = None
    cluster_spec: dict = None
    cluster_instance: dict = None
    overriding_parameters: dict = None
    start_time: int = None
    setup_duration: int = None
    execution_duration: int = None
    cleanup_duration: int = None
    trigger: dict = None
    run_name: str = None
    run_page_url: str = None
    run_type: str = None

@dataclass
class DatabricksRunList(DataClass):
    runs: List[DatabricksRun]
    has_more: bool

@dataclass
class DatabircksNotebookOutput(DataClass):
    result: str = None
    truncated: bool = False


@dataclass
class DatabricksRunOutput(DataClass):
    notebook_output: DatabircksNotebookOutput = None
    error: str = None
    metadata: DatabricksRun = None

    @classmethod
    def from_dict(cls, d: dict) -> 'DatabricksRunOutput':
        notebook_output = d.get('notebook_output', None)
        if notebook_output != None:
            d['notebook_output'] = DatabircksNotebookOutput.from_dict(notebook_output)
        metadata = d.get('metadata', None)
        if metadata != None:
            d['metadata'] = DatabricksRun.from_dict(metadata)
        return DatabricksRunOutput(**d)

@dataclass
class DatabricksViewItem(DataClass):
    content: str = None
    name: str = None
    type: str = None

@dataclass
class DatabricksViewList(DataClass):
    views: List[DatabricksViewItem]

    @classmethod
    def from_dict(cls, d:dict) -> 'DatabricksViewList':
        views = d.get('views', [])
        d['views'] = [DatabricksViewItem.from_dict(v) for v in views]
        return cls(**d)

    @property
    def notebooks(self) -> List[DatabricksViewItem]:
        return [v for v in self.views if v.type == 'NOTEBOOK']



import os
import sys
import base64
import json
import click
from ..connection import connect
from ..common import DatabricksLinkException


global _dbc


@click.group()
@click.option('--bearer-token', '-t', help='Bearer token. Default to DATABRICKS_BEARER_TOKEN environment variable')
@click.option('--url', '-u', help='Databricks URL. Default to DATABRICKS_URL environment variable')
@click.option('--cluster-id', '-c', help='Databricks cluster ID. Default to DATABRICKS_CLUSTER_ID environment variable')
@click.option('-v', help='Verbose 1')
@click.option('-vv', help='Verbose 2')
@click.option('-vvv', help='Verbose 3')
def cli(bearer_token, url, cluster_id, v, vv, vvv):
    global _dbc

    bearer_token = bearer_token or os.environ.get('DATABRICKS_BEARER_TOKEN', None)
    assert bearer_token, 'Bearer token is not provided'
    url = url or os.environ.get('DATABRICKS_URL')
    cluster_id = cluster_id or os.environ.get('DATABRICKS_CLUSTER_ID')

    _dbc = connect(bearer_token, url=url, cluster_id=cluster_id)

@cli.group(help='Databricks workspace commands')
def workspace():
    pass


@workspace.command(help='List Databricks workspace item(s)')
@click.argument('path', default='/', required=False)
@click.option('--json-indent', '-i', help='Number of spaces to use for JSON output indentation.')
def ls(path,json_indent):
    global _dbc
    path = path if path.startswith('/') else '/' + path
    json_indent = int(json_indent) if isinstance(json_indent, str) and json_indent.isnumeric() else json_indent
    try:
        print(json.dumps(_dbc.workspace.ls(path), indent=json_indent, default=lambda o: dict(o)))
    except DatabricksLinkException as exc:
        sys.stderr.write("{}\n".format(str(exc)))
    except Exception as exc:
        raise exc


@workspace.command(help='Export Databricks workspace items')
@click.option('--format','-f', default='SOURCE', help='Export format: DBC, HTML, JUPYTER, SOURCE')
@click.option('--output', '-o', help='Output to a file or directory.')
@click.argument('path', required=False, default='/')
def export(path, output, format):
    def save(item_path, content, item=None):
        languages = ['PYTHON', 'SCALA', 'R', 'SQL']
        extensions = ['.py', '.scala', '.r', '.sql']
        if not output or output == '-':
            print(content)
            return
        if is_recursive:
            rel_path = item_path[len(path):]
            to_path = '{}/{}'.format(output, rel_path)
            try:
                ext = extensions[languages.index(item.language)]
                if not to_path.endswith(ext):
                    to_path = to_path + ext
            except ValueError:
                ext = ''
            to_path = to_path
        else:
            to_path = output
        abs_path = os.path.abspath(to_path)
        dirname = os.path.dirname(abs_path)
        os.makedirs(dirname, exist_ok=True)
        with open(abs_path, 'wb') as fh:
            fh.write(content)

    def export_item(item):
        if item.is_directory:
            export_dir(item.path)
        elif item.is_notebook:
            export_notebook(item)

    def export_dbc(path):
        save(path, _dbc.workspace.export(path, format))

    def export_notebook(item):
        save(item.path, _dbc.workspace.export(item.path, format), item)

    def export_dir(path):
        items = _dbc.workspace.ls(path)
        for item in items:
            export_item(item)


    global _dbc
    path = path if len(path) and path[0] == '/' else '/' + path
    format = format.upper()

    try:
        is_recursive = True
        if format == 'DBC':
            is_recursive = False
            export_dbc(path)
        elif path == '/':
            export_dir(path)
        else:
            item = _dbc.workspace.ls(path)[0]
            if item.path == path and not item.is_directory:
                is_recursive = False
                export_item(item)
            else:
                is_recursive = True
                export_dir(path)
    except DatabricksLinkException as exc:
        sys.stderr.write("{}\n".format(str(exc)))
    except Exception as exc:
        raise exc

@cli.group(help='Databricks DBFS commands')
def dbfs():
    pass


@dbfs.command(help='List DBFS item(s)')
@click.argument('path', required=False, default='/', type=str)
@click.option('--json-indent', help='Number of spaces to use for JSON output indentation.')
def ls(path, json_indent):
    global _dbc
    path = path if len(path) and path[0] == '/' else '/' + path
    json_indent = int(json_indent) if isinstance(json_indent, str) and json_indent.isnumeric() else json_indent

    try:
        result = _dbc.dbfs.ls(path)
        print(json.dumps(result, indent=json_indent, default=lambda o: dict(o)))
    except DatabricksLinkException as exc:
        sys.stderr.write("{}\n".format(str(exc)))
    except Exception as exc:
        raise exc



@dbfs.command(help='Download a file from DBFS')
@click.option('--chunk-size', '-c', default=1024*1024, help='Download chunk size')
@click.option('--output', '-o', default='-', help='Output file. Default is "-", writing to stdout')
@click.argument('path')
def get(path, output, chunk_size):
    def get_item(item, recursive):
        if item.is_file:
            file_path = output
            if recursive and output != '-':
                rel_path = item.path[len(path):]
                file_path = os.path.abspath('{}/{}'.format(output, rel_path))
                os.makedirs(os.path.dirname(file_path), exist_ok=True)
            get_file(item.path, file_path)
        else:
            for dir_item in _dbc.dbfs.ls(item.path):
                get_item(dir_item, recursive)

    def get_file(path, output):
        fh = sys.stdout.buffer if output == '-' else open(output, 'wb')
        try:
            offset = 0
            while (True):
                this_read = _dbc.dbfs.read(
                        path, 
                        offset=offset,
                        length=chunk_size,
                        decoded=False)
                if not this_read['bytes_read']:
                    break
                offset += this_read['bytes_read']
                content = base64.b64decode(this_read['data'])
                fh.write(content)
        finally:
            if fh is not sys.stdout.buffer:
                fh.close()


    global _dbc
    path = path if len(path) and path[0] == '/' else '/' + path
    
    try:
        item = _dbc.dbfs.info(path)
        get_item(item, item.is_dir)
    except DatabricksLinkException as exc:
        sys.stderr.write("{}\n".format(str(exc)))
    except Exception as exc:
        raise exc

@dbfs.command(help='Remove a file or directory from DBFS')
@click.option('--recursive', '-r', is_flag=True, help='Remove recursively')
@click.argument('path')
def rm(path, recursive):
    global _dbc
    path = path if len(path) and path[0] == '/' else '/' + path

    try:
        _dbc.dbfs.delete(path, recursive)
    except DatabricksLinkException as exc:
        sys.stderr.write("{}\n".format(str(exc)))
    except Exception as exc:
        raise exc

@dbfs.command(help='Remove a file or directory from DBFS')
@click.argument('path')
def mkdirs(path):
    global _dbc
    path = path if len(path) and path[0] == '/' else '/' + path

    try:
        _dbc.dbfs.mkdirs(path)
    except DatabricksLinkException as exc:
        sys.stderr.write("{}\n".format(str(exc)))
    except Exception as exc:
        raise exc

# ######################## Jobs ###########################

@cli.group(help='Databricks jobs commands')
def jobs():
    pass


@jobs.command(help='List databricks jobs')
@click.argument('path', required=False, default='/', type=str)
@click.option('--json-indent', help='Number of spaces to use for JSON output indentation.')
def ls(path, json_indent):
    global _dbc
    path = path if path.startswith('/') else '/' + path
    json_indent = int(json_indent) if isinstance(json_indent, str) and json_indent.isnumeric() else json_indent

    try:
        result = _dbc.jobs.list(path)
        print(json.dumps(result, indent=json_indent, default=lambda o: dict(o)))
    except DatabricksLinkException as exc:
        sys.stderr.write("{}\n".format(str(exc)))
    except Exception as exc:
        raise exc


# ######################## Runs ###########################

@cli.group(help='Databricks runs commands')
def runs():
    pass


@runs.command()
@click.argument('job-id', required=False, type=str)
@click.option('--completed-only', is_flag=True, help='List only completed runs.')
@click.option('--active-only', is_flag=True, help='List only active runs.')
@click.option('--offset', help='Page offset.')
@click.option('--limit', help='Page length.')
@click.option('--json-indent', '-i', help='Number of spaces to use for JSON output indentation.')
def ls(job_id=None, completed_only=False, active_only=False, offset=None, limit=None, json_indent=None):
    """
    List runs from most recently started to least.

    See: https://docs.databricks.com/dev-tools/api/latest/jobs.html#runs-list
    """
    global _dbc
    json_indent = int(json_indent) if isinstance(json_indent, str) and json_indent.isnumeric() else json_indent

    try:
        result = _dbc.jobs.runs.ls(job_id=job_id, completed_only=completed_only, active_only=active_only,
                offset=offset, limit=limit)
        print(json.dumps(result, indent=json_indent, default=lambda o: dict(o)))
    except DatabricksLinkException as exc:
        sys.stderr.write("{}\n".format(str(exc)))
    except Exception as exc:
        raise exc

@runs.command()
@click.argument('path', required=True, type=str)
@click.option('--run-name', '-n', help='Run name')
@click.option('--parameters', '-p', type=str, help='Parameters JSON object.')
@click.option('--cluster-id', '-c', help='Cluster ID. Default from DATABRICKS_CLUSTER_ID environment variable.')
@click.option('--json-indent', help='Number of spaces to use for JSON output indentation.')
def submit(path, run_name=None, parameters=None, cluster_id=None, json_indent=None):
    '''
    Submit a one-time run. Doesn’t require a Databricks job to be created.

    See: https://docs.databricks.com/dev-tools/api/latest/jobs.html#runs-submit
    '''
    global _dbc
    path = path if path.startswith('/') else '/' + path
    json_indent = int(json_indent) if isinstance(json_indent, str) and json_indent.isnumeric() else json_indent

    if parameters:
        if parameters.startswith('@'):
            with open(parameters[1:], 'r', encoding='utf8') as f:
                parameters = f.read()
        params = json.loads(parameters)
    else:
        params = None

    try:
        result = _dbc.jobs.runs.submit_notebook(path=path, params=params, run_name=run_name)
        print(json.dumps({'run_id': result}, indent=json_indent, default=lambda o: dict(o)))
    except DatabricksLinkException as exc:
        sys.stderr.write("{}\n".format(str(exc)))
    except Exception as exc:
        raise exc


@runs.command()
@click.argument('run-id', required=True, type=int)
@click.option('--json-indent', '-i', help='Number of spaces or string to use for JSON output indentation.')
# https://docs.databricks.com/dev-tools/api/latest/jobs.html#runs-get
def get(run_id=None, json_indent=None):
    '''
    Retrieve the metadata of a run.

    See: https://docs.databricks.com/dev-tools/api/latest/jobs.html#runs-get
    '''
    global _dbc
    json_indent = int(json_indent) if isinstance(json_indent, str) and json_indent.isnumeric() else json_indent

    try:
        result = _dbc.jobs.runs.get(run_id=run_id)
        print(json.dumps(result, indent=json_indent, default=lambda o: dict(o)))
    except DatabricksLinkException as exc:
        sys.stderr.write("{}\n".format(str(exc)))
    except Exception as exc:
        raise exc

@runs.command()
@click.argument('run-id', required=True, type=int)
@click.option('--content-only', '-c', is_flag=True, help="Content only.")
@click.option('--json-indent', '-i', help='Number of spaces or string to use for JSON output indentation.')
# https://docs.databricks.com/dev-tools/api/latest/jobs.html#runs-get
def export(run_id=None, content_only=False, json_indent=None):
    '''
    Export and retrieve the job run task.

    See: https://docs.databricks.com/dev-tools/api/latest/jobs.html#runs-export
    '''
    global _dbc
    json_indent = int(json_indent) if isinstance(json_indent, str) and json_indent.isnumeric() else json_indent

    try:
        result = _dbc.jobs.runs.export(run_id=run_id)
        if content_only:
            for item_otput in result.views:
                print(item_otput.content)
        else:
            print(json.dumps(result.views, indent=json_indent, default=lambda o: dict(o)))
    except DatabricksLinkException as exc:
        sys.stderr.write("{}\n".format(str(exc)))
    except Exception as exc:
        raise exc


@runs.command()
@click.argument('run-id', required=True, type=int)
@click.option('--result-only', '-r', is_flag=True, help="Content only.")
@click.option('--json-indent', '-i', help='Number of spaces or string to use for JSON output indentation.')
# https://docs.databricks.com/dev-tools/api/latest/jobs.html#runs-get
def get_output(run_id=None, result_only=False, json_indent=None):
    '''
    Retrieve the output of a run.

    When a notebook task returns a value through the dbutils.notebook.exit() call, 
    you can use this endpoint to retrieve that value. Databricks restricts this API 
    to return the first 5 MB of the output. For returning a larger result, you can 
    store job results in a cloud storage service.

    See: https://docs.databricks.com/dev-tools/api/latest/jobs.html#runs-export
    '''
    global _dbc
    json_indent = int(json_indent) if isinstance(json_indent, str) and json_indent.isnumeric() else json_indent

    try:
        result = _dbc.jobs.runs.get_output(run_id=run_id)
        if result_only:
            print(result.notebook_output.result)
        else:
            print(json.dumps(result, indent=json_indent, default=lambda o: dict(o)))
    except DatabricksLinkException as exc:
        sys.stderr.write("{}\n".format(str(exc)))
    except Exception as exc:
        raise exc




from typing import List, Callable
import pyspark.sql as ps
import pyspark.sql.types as pst
import pyspark



def as_list(input_df:ps.DataFrame, deep:bool=False, 
            order_by:Callable[[object, object],object]=None) -> List[dict]:
    """Convert Spark DataFrame to Python list of dictionaries.
    
    The DataFrame rows are converted into dictionaries and collected into 
    a list of dictionary. Optionally structure fields of type Row type are 
    further recursively converted to dictionaries.
    
    By default rows in the output list are not ordered. Optionally a key
    function could be supplied so that the output list is ordered.
    Ordered list is useful, for example, when testing to compare actual 
    data against expected data.
    
    Parameters:
      input_df (pyspark.sql.DataFrame): Spark DataFrame to be converted.
      
      deep (bool): (Default=False) Whether Row to dictionary conversion should be deep.
      
      order_by (callable): (Default=None) Key function to order by. If provided, the result 
                           list will be sorted using the provided key function.
    Returns:
      list: A list of records, converted to dictionary.
    """
    def row_as_dict(r):
      d = r.asDict()
      if deep:
        for k in d:
          if isinstance(d[k], pst.Row):
            d[k] = row_as_dict(d[k])
      return d
    
    if order_by:
      assert callable(order_by), "order_by must be callable"
    if isinstance(input_df, ps.DataFrame):
        result = [row_as_dict(r) for r in input_df.collect()]
    else:
        result = input_df
    if order_by and isinstance(result, list):
        result.sort(key=order_by)
    return result




from operator import itemgetter
import pytest
import pyspark.sql.types as pst
from pydbr.pyspark.functions import *


def test_as_list_converts_rows(spark):
    """Given a dataframe, 
       When as_list is called,
       Then result is a list of dictionaries"""
    input = [ (1, 'John') ]
    input_df = spark.createDataFrame(input, ['id', 'first_name'])

    actual = as_list(input_df, False)
    expect = [ dict(id=1, first_name='John')]

    assert actual == expect


def test_as_list_shallow_doesnt_convert_fields(spark):
    """Given a dataframe with nested structure fields, 
       When as_list with shallow=True is called,
       Then result is a list of dictionaries with structure field not converted"""
    input = [ (1, pst.Row(first_name='John')) ]
    input_df = spark.createDataFrame(input, ['id', 'person'])

    actual = as_list(input_df, False)
    expect = [ dict(id=1, person=pst.Row(first_name='John'))]

    assert actual == expect


def test_as_list_deep_convert_fields(spark):
    """Given a dataframe with nested structure fields, 
       When as_list with shallow=True is called,
       Then result is a list of dictionaries with structure field not converted"""
    input = [ (1, pst.Row(first_name='John')) ]
    input_df = spark.createDataFrame(input, ['id', 'person'])

    actual = as_list(input_df, False)
    expect = [ dict(id=1, person=pst.Row(first_name='John'))]

    assert actual == expect


def test_as_list_with_order_by_sorts_rows(spark):
    """Given unsorted dataframe with two rows,
       When as_list with order_by is called,
       Then result list is ordered"""
    
    input = [(9, 'Zztop'),
             (1, 'John')]
    input_df = spark.createDataFrame(input, ['id', 'name'])
    
    expect = [dict(id=1, name='John'),
              dict(id=9, name='Zztop')]
    
    actual = as_list(input_df, order_by=itemgetter('id'))
    assert actual == expect



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pydbr",
    version="0.0.6",
    author="Ivan Georgiev",
    #author_email="ivan.georgiev",
    description="Databricks client SDK with command line client for Databricks REST APIs",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ivangeorgiev/pydbr",
    packages=setuptools.find_packages(),
    install_requires=[
        'click',
        'requests'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points='''
        [console_scripts]
        pydbr=pydbr.cli:cli
    ''',
    python_requires='>=3.6',
)



import requests
from collections import namedtuple
import base64

from .connection import connect

DEFAULT_URL = 'https://westeurope.azuredatabricks.net/'

_bearer_token = None
_cluster_id = None
_url = None

DBFS_FileInfo = namedtuple('DBFS_FileInfo', 'path,is_dir,file_size')

def _get_headers():
    global _bearer_token
    return dict(
       Authorization = f'Bearer {_bearer_token}'
    )

def _get_url(endpoint):
    global _url
    url = f'{_url}/api/2.0/{endpoint}'
    return url

def _get_cluster_id(cluster_id=None):
    global _cluster_id
    return (cluster_id or _cluster_id)

def connect(bearer_token, url=None, cluster_id=None):
    global _bearer_token, _url, _cluster_id
    _bearer_token = bearer_token
    _url = url or DEFAULT_URL
    _cluster_id = cluster_id


def dbfs_ls(path):
    global _url, _bearer_token

    url = f'{_url}/api/2.0/dbfs/list'
    params = dict(path=path)
    headers = _get_headers()

    response_obj = requests.get(url,params=params,
            headers=headers)

    response = response_obj.json()
    if response.get('error_code', None) == 'RESOURCE_DOES_NOT_EXIST':
        raise FileNotFoundError(response['message'])
    elif response.get('error_code', None):
        raise Exception(response['error_code'] + ' ' + response['message'])
    return [DBFS_FileInfo(**f) for f in response['files']]

def dbfs_mkdirs(path):
    global _url, _bearer_token

    url = f'{_url}/api/2.0/dbfs/mkdirs'
    params = dict(path=path)
    headers = _get_headers()

    response_obj = requests.post(url,params=params,
            headers=headers)

    response = response_obj.json()
    if response.get('error_code', None):
        raise Exception(response['error_code'] + ' ' + response['message'])

def dbfs_delete(path, recursive=False):

    url = _get_url('dbfs/delete')
    headers = _get_headers()
    params = dict(
        path=path,
        recursive=str(recursive).lower(),
        )

    response_obj = requests.post(
            url,
            params=params,
            headers=headers,
            )

    response = response_obj.json()
    if response.get('error_code', None):
        raise Exception(response['error_code'] + ' ' + response['message'])


def dbfs_read(path, offset=None, length=None, decoded=True):
    offset = offset or 0
    length = length or 1048576

    response_obj = requests.get(
        url=_get_url('dbfs/read'),
        params=dict(path=path,offset=offset,length=length),
        headers=_get_headers()
    )
    response = response_obj.json()
    if response.get('error_code', None):
        raise Exception(response['error_code'] + ' ' + response['message'])
    if decoded:
        response = base64.b64decode(response['data'])
    return response

def dbfs_read_all(path, chunk_size=None) -> bytes:
    chunk_size = chunk_size or 1048576
    content = b''
    offset = 0
    while (True):
        this_read = dbfs_read(
                path, 
                offset=offset,
                length=chunk_size,
                decoded=False)
        if not this_read['bytes_read']:
            break
        offset += this_read['bytes_read']
        content += base64.b64decode(this_read['data'])
    return content


def run_submit_notebook(path, params=None, run_name=None, cluster_id=None):
    cluster_id = _get_cluster_id(cluster_id)
    assert cluster_id, f"cluster_id not specified. Set cluster_id with connect or pass as parameter"

    params = params or {}

    run_name = None or "My Notebook Run"
    r = dict(
        run_name=run_name,
        existing_cluster_id=cluster_id,
        libraries=[],
        notebook_task=dict(
            notebook_path=path,
            base_parameters=params
        ),
    )


    response_obj = requests.post(
        url=_get_url('jobs/runs/submit'),
        headers=_get_headers(),
        json=r
    )

    response = response_obj.json()
    if response.get('error_code', None):
        raise Exception(response['error_code'] + ' ' + response['message'])
    return response['run_id']

def run_get(run_id):
    response_obj = requests.get(
        url=_get_url('jobs/runs/get'),
        params=dict(run_id=run_id),
        headers=_get_headers()
    )
    response = response_obj.json()
    if response.get('error_code', None):
        raise Exception(response['error_code'] + ' ' + response['message'])
    return response


def run_get_output(run_id):
    response_obj = requests.get(
        url=_get_url('jobs/runs/get-output'),
        params=dict(run_id=run_id),
        headers=_get_headers()
    )
    response = response_obj.json()
    if response.get('error_code', None):
        raise Exception(response['error_code'] + ' ' + response['message'])
    return response

def run_export(run_id):
    response_obj = requests.get(
        url=_get_url('jobs/runs/export'),
        params=dict(run_id=run_id),
        headers=_get_headers()
    )
    response = response_obj.json()
    if response.get('error_code', None):
        raise Exception(response['error_code'] + ' ' + response['message'])
    return response


def run_list(job_id=None, offset=None, limit=None,
        completed_only=False, active_only=False):
    assert not (completed_only and active_only), "Only one of completed_only or active_only could be True"
    params = dict()
    if job_id:
        params['job_id'] = job_id
    if offset:
        params['offset'] = offset
    if limit:
        params['limit'] = limit
    if completed_only:
        params['completed_only'] = 'true'
    if active_only:
        params['active_only'] = 'true'

    response_obj = requests.get(
        url=_get_url('jobs/runs/list'),
        params=params,
        headers=_get_headers()
    )
    response = response_obj.json()
    if response.get('error_code', None):
        raise Exception(response['error_code'] + ' ' + response['message'])
    return response



import base64
from typing import List
from .common import Api, DatabricksLinkException, ERR_RESOURCE_DOES_NOT_EXIST
from .databricks_data_classes import *

class DBFS(Api):
    def __init__(self, link):
        super().__init__(link, path='dbfs')

    def list(self, path=None) -> List[DatabricksFileInfo]:
        "List DBFS items at given path. Alias for ls"
        get_result = self.link.get(
            self.path('list'),
            params=dict(path=(path or '/')))
        files = get_result.get('files', [])
        result = [DatabricksFileInfo(**f) for f in files]
        return result

    def info(self, path=None) -> DatabricksFileInfo:
        """Get information for DBFS item."""
        response = self.link.get(
            self.path('get-status'),
            params=dict(path=(path or '/')))
        result = DatabricksFileInfo(**response)
        return result

    def ls(self, path=None) -> List[DatabricksFileInfo]:
        """List DBFS items at given path."""
        return self.list(path)

    def exists(self, path) -> bool:
        """Check if given DBFS item exists."""
        try:
            self.list(path)
            result = True
        except DatabricksLinkException as exc:
            if exc.error_code == ERR_RESOURCE_DOES_NOT_EXIST:
                result = False
        return result


    def read_raw(self, path, offset=None, length=None):
        """Read a chunk from a DBFS file and return the API response."""
        offset = offset or 0
        length = length or 1048576

        response = self.link.get(
            self.path('read'),
            params=dict(path=path,offset=offset,length=length),)
        return response

    def read(self, path, offset=None, length=None) -> bytes:
        """Read a chunk from a DBFS file and return the bytes."""
        response = self.read_raw(path, offset, length)
        return base64.b64decode(response['data'])


    def read_all(self, path, chunk_size=None) -> bytes:
        """Read entire file from DBFS and return they bytes."""
        chunk_size = chunk_size or 1048576
        content = b''
        offset = 0
        while (True):
            this_read = self.read_raw(
                    path, 
                    offset=offset,
                    length=chunk_size)
            if not this_read['bytes_read']:
                break
            offset += this_read['bytes_read']
            content += base64.b64decode(this_read['data'])
        return content

    def mkdirs(self, path):
        """Create missing DBFS directories recursively."""
        response = self.link.post(
            self.path('mkdirs'),
            json=dict(path=path))
        return response

    def delete(self, path, recursive=False):
        """Remove DBFS item. Alias for rm"""
        response = self.link.post(
            self.path('delete'),
            json=dict(path=path,
                      recursive=str(recursive).lower()))
        return response

    def rm(self, path, recursive=False):
        """Remove DBFS items."""
        return self.delete(path, recursive)
